/*
 * Copyright (c) 1999 by David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 *
 * Most simple built-in commands are here.
 */

#include "message.h"
#include "sash.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mount.h>
#include <signal.h>
#include <pwd.h>
#include <grp.h>
#include <utime.h>
#ifndef	dev_t
#define	dev_t	dev_t
#endif
#include <linux/loop.h>


void
do_echo(int argc, const char *const *argv)
{
	BOOL	first;

	first = TRUE;

	while (argc-- > 1)
	{
		if (!first)
			fputc(' ', stdout);

		first = FALSE;
		fputs(*(++argv), stdout);
	}

	fputc('\n', stdout);
}


void
do_pwd(int argc, const char *const *argv)
{
	char	buf[PATH_LEN];

	if ( !getcwd(buf, PATH_LEN) )
	{
		errmsg( "pwd: cannot get current directory" );
		return;
	}

	printf("%s\n", buf);
}


void
do_cd(int argc, const char *const *argv)
{
	const char *	path;

	if (argc > 1)
		path = argv[1];
	else
	{
		path = getenv("HOME");

		if ( !path )
		{
			errmsg( "cd: HOME not set" );
			return;
		}
	}

	if (chdir(path) < 0)
		errmsg( "cd: chdir: %s: %s", path, strerror(errno) );
}


void
do_mkdir(int argc, const char *const *argv)
{
	while (argc-- > 1)
	{
		if (mkdir(argv[1], 0777) < 0)
			errmsg( "mkdir: %s: %s", argv[1], strerror(errno) );

		++argv;
	}
}


void
do_mknod(int argc, const char *const *argv)
{
	int	mode = 0666, major = 0, minor = 0;
	const	char	*cp = argv[3];

	if (strcmp(argv[2], "b") == 0)
		mode |= S_IFBLK;
	else if (strcmp(argv[2], "c") == 0)
		mode |= S_IFCHR;
	else
	{
		errmsg( "mknod: bad device type: %s", argv[2] );
		return;
	}

	while (isDecimal(*cp))
		major = major * 10 + *cp++ - '0';

	if (*cp || (major < 0) || (major > 255))
	{
		errmsg( "mknod: bad major number: %d", major );
		return;
	}

	cp = argv[4];

	while (isDecimal(*cp))
		minor = minor * 10 + *cp++ - '0';

	if (*cp || (minor < 0) || (minor > 255))
	{
		errmsg( "mknod: bad minor number: %d", minor );
		return;
	}

	if (mknod(argv[1], mode, major * 256 + minor) < 0)
		errmsg( "mknod: %s: %s", argv[1], strerror(errno) );
}


void
do_rmdir(int argc, const char *const *argv)
{
	while (argc-- > 1)
	{
		if (rmdir(argv[1]) < 0)
			errmsg( "rmdir: %s: %s", argv[1], strerror(errno) );

		++argv;
	}
}


void
do_sync(int argc, const char *const *argv)
{
	sync();
}


void
do_rm(int argc, const char *const *argv)
{
	while (argc-- > 1)
	{
		if (unlink(argv[1]) < 0)
			errmsg( "rm: %s: %s", argv[1], strerror(errno) );

		++argv;
	}
}


void
do_chmod(int argc, const char *const *argv)
{
	const char *	cp;
	int		mode;

	mode = 0;
	cp = argv[1];

	while (isOctal(*cp))
		mode = mode * 8 + (*cp++ - '0');

	if (*cp)
	{
		errmsg( "chmod: mode must be octal" );
		return;
	}

	--argc;
	++argv;

	while (argc-- > 1)
	{
		if (chmod(argv[1], mode) < 0)
			errmsg( "chmod: %s: %s", argv[1], strerror(errno) );

		++argv;
	}
}


void
do_chown(int argc, const char *const *argv)
{
	const char *	cp;
	int		uid;
	struct passwd *	pwd;
	struct stat	statBuf;

	cp = argv[1];

	if (isDecimal(*cp))
	{
		uid = 0;

		while (isDecimal(*cp))
			uid = uid * 10 + (*cp++ - '0');

		if (*cp)
		{
			errmsg( "chown: bad uid value: %s", argv[1] );
			return;
		}
	} else {
		pwd = getpwnam(cp);

		if ( !pwd )
		{
			errmsg( "chown: unknown user name: %s", cp );
			return;
		}

		uid = pwd->pw_uid;
	}

	--argc;
	++argv;

	while (argc-- > 1)
	{
		++argv;

		if ((stat(*argv, &statBuf) < 0) ||
			(chown(*argv, uid, statBuf.st_gid) < 0))
		{
			errmsg( "chown: %s: %s", *argv, strerror(errno) );
		}
	}
}


void
do_chgrp(int argc, const char *const *argv)
{
	const char *	cp;
	int		gid;
	struct group *	grp;
	struct stat	statBuf;

	cp = argv[1];

	if (isDecimal(*cp))
	{
		gid = 0;

		while (isDecimal(*cp))
			gid = gid * 10 + (*cp++ - '0');

		if (*cp)
		{
			errmsg( "chgrp: bad gid value: %s", argv[1] );
			return;
		}
	}
	else
	{
		grp = getgrnam(cp);

		if ( !grp )
		{
			errmsg( "chgrp: unknown group name: %s", cp );
			return;
		}

		gid = grp->gr_gid;
	}

	--argc;
	++argv;

	while (argc-- > 1)
	{
		++argv;

		if ((stat(*argv, &statBuf) < 0) ||
			(chown(*argv, statBuf.st_uid, gid) < 0))
		{
			errmsg( "chgrp: %s: %s", *argv, strerror(errno) );
		}
	}
}


void
do_touch(int argc, const char *const *argv)
{
	const char *	name;
	int		fd;
	struct utimbuf	now;

	time(&now.actime);
	now.modtime = now.actime;

	while (argc-- > 1)
	{
		name = *(++argv);

		fd = open(name, O_CREAT | O_WRONLY | O_EXCL, 0666);

		if (fd >= 0)
		{
			close(fd);
			continue;
		}

		if (utime(name, &now) < 0)
			errmsg( "touch: %s: %s", name, strerror(errno) );
	}
}


void
do_mv(int argc, const char *const *argv)
{
	const	char	*srcName;
	const	char	*destName;
	const	char	*lastArg = argv[argc - 1];
	BOOL	dirFlag = isDirectory(lastArg);

	if ((argc > 3) && !dirFlag)
	{
		errmsg( "mv: %s: when moving multiple files, last argument must be a directory", lastArg );
		return;
	}

	while (!intFlag && (argc-- > 2))
	{
		srcName = *(++argv);

		if (access(srcName, 0) < 0)
		{
			errmsg( "mv: %s: %s", srcName, strerror(errno) );
			continue;
		}

		destName = lastArg;

		if (dirFlag)
			destName = buildName(destName, srcName);

		if (rename(srcName, destName) >= 0)
			continue;

		if (errno != EXDEV)
		{
			errmsg( "mv: %s: %s", destName, strerror(errno) );
			continue;
		}

		if (!copyFile(srcName, destName, TRUE))
			continue;

		if (unlink(srcName) < 0)
			errmsg( "mv: error unlinking %s: %s", srcName, strerror(errno) );
	}
}


void
do_ln(int argc, const char *const *argv)
{
	const	char *	srcName;
	const	char *	destName;
	const	char *	lastArg = argv[argc - 1];
	BOOL		dirFlag = isDirectory(lastArg), sym = FALSE;

	if (argv[1][0] == '-')
	{
		if (strcmp(argv[1], "-s"))
		{
			errmsg( "ln: unknown option: %s", argv[1] );
			return;
		}

		sym = TRUE;
		--argc;
		++argv;
	}

	if ((argc > 3) && !dirFlag)
	{
		errmsg( "ln: %s: when making multiple links, last argument must be a directory", lastArg );
		return;
	}

	while (--argc > 1)
	{
		srcName = *(++argv);

		if ( !sym && access(srcName, 0) < 0)
		{
			errmsg( "ln: %s: %s", srcName, strerror(errno) );
			continue;
		}

		destName = lastArg;

		if (dirFlag)
			destName = buildName(destName, srcName);

		if ( (sym ? symlink(srcName, destName) : link(srcName, destName)) < 0 )
			errmsg( "%s: %s -> %s: %s", (sym ? "symlink" : "ln"),
				srcName, destName, strerror(errno) );
	}
}


void
do_cp(int argc, const char *const *argv)
{
	const char *	srcName;
	const char *	destName;
	const char *	lastArg;
	BOOL		dirFlag;

	lastArg = argv[argc - 1];

	dirFlag = isDirectory(lastArg);

	if ((argc > 3) && !dirFlag)
	{
		errmsg( "cp: %s: when copying multiple files, last argument must be a directory", lastArg );
		return;
	}

	while (!intFlag && (argc-- > 2))
	{
		srcName = *(++argv);
		destName = lastArg;

		if (dirFlag)
			destName = buildName(destName, srcName);

		(void) copyFile(srcName, destName, FALSE);
	}
}


void
do_mount(int argc, const char *const *argv)
{
	const char *	str;
	const char *	type;
	int		flags;

	--argc;
	++argv;
	type = "ext2";
	flags = MS_MGC_VAL;

	while ((argc > 0) && (**argv == '-'))
	{
		--argc;
		str = *argv++;

		while (*++str) switch (*str)
		{
			case 't':
				if ((argc <= 0) || (**argv == '-'))
				{
					errmsg( "mount: missing file system type" );
					return;
				}

				type = *argv++;
				--argc;
				break;

			case 'r':
				flags |= MS_RDONLY;
				break;

			case 'm':
				flags |= MS_REMOUNT;
				break;

			default:
				errmsg( "mount: unknown option: %s", str );
				return;
		}
	}

	if (argc != 2)
	{
		errmsg( "mount: wrong number of arguments for mount" );

		return;
	}

	if (mount(argv[0], argv[1], type, flags, 0) < 0)
		errmsg( "mount: %s", strerror(errno) );
}


void
do_umount(int argc, const char *const *argv)
{
	if (umount(argv[1]) < 0)
		errmsg( "umount: %s: %s", argv[1], strerror(errno) );
}


void
do_cmp(int argc, const char *const *argv)
{
	int		fd1;
	int		fd2;
	int		cc1;
	int		cc2;
	long		pos;
	const char *	bp1;
	const char *	bp2;
	char		buf1[BUF_SIZE];
	char		buf2[BUF_SIZE];
	struct	stat	statBuf1;
	struct	stat	statBuf2;

	if (stat(argv[1], &statBuf1) < 0)
	{
		errmsg( "cmp: %s: %s", argv[1], strerror(errno) );
		return;
	}

	if (stat(argv[2], &statBuf2) < 0)
	{
		errmsg( "cmp: %s: %s", argv[2], strerror(errno) );
		return;
	}

	if ((statBuf1.st_dev == statBuf2.st_dev) &&
		(statBuf1.st_ino == statBuf2.st_ino))
	{
		puts( "Files are links to each other" );
		return;
	}

	if (statBuf1.st_size != statBuf2.st_size)
	{
		puts( "Files are different sizes" );
		return;
	}

	fd1 = open(argv[1], O_RDONLY);

	if (fd1 < 0)
	{
		errmsg( "cmp: error opening %s: %s", argv[1], strerror(errno) );
		return;
	}

	fd2 = open(argv[2], O_RDONLY);

	if (fd2 < 0)
	{
		errmsg( "cmp: error opening %s: %s", argv[2], strerror(errno) );
		close(fd1);
		return;
	}

	pos = 0;

	for(;;)
	{
		if (intFlag)
			goto closefiles;

		cc1 = read(fd1, buf1, sizeof(buf1));

		if (cc1 < 0)
		{
			errmsg( "cmp: error reading %s: %s", argv[1], strerror(errno) );
			goto closefiles;
		}

		cc2 = read(fd2, buf2, sizeof(buf2));

		if (cc2 < 0)
		{
			errmsg( "cmp: error reading %s: %s", argv[2], strerror(errno) );
			goto closefiles;
		}

		if ( !cc1 && !cc2 )
		{
			puts( "Files are identical" );
			goto closefiles;
		}

		if (cc1 < cc2)
		{
			puts( "First file is shorter than second" );
			goto closefiles;
		}

		if (cc1 > cc2)
		{
			puts( "Second file is shorter than first" );
			goto closefiles;
		}

		if (memcmp(buf1, buf2, cc1) == 0)
		{
			pos += cc1;

			continue;
		}

		bp1 = buf1;
		bp2 = buf2;

		while (*bp1++ == *bp2++)
			pos++;

		printf( "Files differ at byte position %ld\n", pos );

		goto closefiles;
	}

closefiles:
	close(fd1);
	close(fd2);
}


void
do_more(int argc, const char *const *argv)
{
	FILE *		fp;
	const char *	name;
	int		ch;
	int		line;
	int		col;
	int		pageLines;
	int		pageColumns;
	char		buf[80];

	/*
	 * Get the width and height of the screen if it is set.
	 * If not, then default it.
	 */
	pageLines = 0;
	pageColumns = 0;

	name = getenv("LINES");

	if (name)
		pageLines = atoi(name);

	name = getenv("COLS");

	if (name)
		pageColumns = atoi(name);

	if (pageLines <= 0)
		pageLines = 24;

	if (pageColumns <= 0)
		pageColumns = 80;

	/*
	 * OK, process each file.
	 */
	while (argc-- > 1)
	{
		name = *(++argv);

		fp = fopen(name, "r");
		if ( !fp )
		{
			errmsg( "more: error opening %s: %s", name, strerror(errno) );
			return;
		}

		printf("<< %s >>\n", name);
		line = 1;
		col = 0;

		while (fp && ((ch = fgetc(fp)) != EOF))
		{
			switch (ch)
			{
				case '\r':
					col = 0;
					break;

				case '\n':
					line++;
					col = 0;
					break;

				case '\t':
					col = ((col + 1) | 0x07) + 1;
					break;

				case '\b':
					if (col > 0)
						col--;
					break;

				default:
					col++;
			}

			putchar(ch);

			if (col >= pageColumns)
			{
				col -= pageColumns;
				line++;
			}

			if (line < pageLines)
				continue;

			if (col > 0)
				putchar('\n');

			printf("--More--");
			fflush(stdout);

			if (intFlag || (read(0, buf, sizeof(buf)) < 0))
			{
				if (fp)
					fclose(fp);

				return;
			}

			ch = buf[0];

			if (ch == ':')
				ch = buf[1];

			switch (ch)
			{
				case 'N':
				case 'n':
					fclose(fp);
					fp = NULL;
					break;

				case 'Q':
				case 'q':
					fclose(fp);

					return;
			}

			col = 0;
			line = 1;
		}

		if (fp)
			fclose(fp);
	}
}


void
do_sum(int argc, const char *const *argv)
{
	const char *	name;
	int		fd;
	int		cc;
	int		ch;
	int		i;
	unsigned long	checksum;
	char		buf[BUF_SIZE];

	--argc;
	++argv;

	while (argc-- > 0)
	{
		name = *argv++;

		fd = open(name, O_RDONLY);

		if (fd < 0)
		{
			errmsg( "sum: error opening %s: %s", name, strerror(errno) );
			continue;
		}

		checksum = 0;

		while ((cc = read(fd, buf, sizeof(buf))) > 0)
		{
			for (i = 0; i < cc; i++)
			{
				ch = buf[i];

				if ((checksum & 0x01) != 0)
					checksum = (checksum >> 1) + 0x8000;
				else
					checksum = (checksum >> 1);

				checksum = (checksum + ch) & 0xffff;
			}
		}

		if (cc < 0)
		{
			errmsg( "sum: error reading %s: %s", name, strerror(errno) );
			(void) close(fd);

			continue;
		}

		(void) close(fd);

		printf("%05lu %s\n", checksum, name);
	}
}


void
do_exit(int argc, const char *const *argv)
{
	if ( getpid() == 1 )
	{
		errmsg( "exit: you are the INIT process!" );
		return;
	}

	exit( argc ? atoi( argv[1] ) : 0 );
}


void
do_setenv(int argc, const char *const *argv)
{
	if ( setenv( argv[1], argv[2], 1 ) < 0 )
		errmsg( "setenv: %s", strerror(errno) );
}


void
do_printenv(int argc, const char *const *argv)
{
	const char *const *	env;
	extern char **	environ;
	int		len;

	env = (const char *const *) environ;

	if (argc == 1)
	{
		while (*env)
			printf("%s\n", *env++);

		return;
	}

	len = strlen(argv[1]);

	while (*env)
	{
		if ((strlen(*env) > len) && (env[0][len] == '=') &&
			(memcmp(argv[1], *env, len) == 0))
		{
			printf("%s\n", &env[0][len+1]);

			return;
		}
		env++;
	}
}


void
do_umask(int argc, const char *const *argv)
{
	const char *	cp;
	int		mask;

	if (argc <= 1)
	{
		mask = umask(0);
		umask(mask);
		printf("%03o\n", mask);

		return;
	}

	mask = 0;
	cp = argv[1];

	while (isOctal(*cp))
		mask = mask * 8 + *cp++ - '0';

	if (*cp || (mask & ~0777))
	{
		errmsg( "umask: bad value: %o", mask );
		return;
	}

	umask(mask);
}

typedef struct
{
	int	no;
	const char *name;
} sig_desc_t;

static sig_desc_t	sig_desc[] = {
	{ SIGHUP, "HUP" },
	{ SIGINT, "INT" },
	{ SIGQUIT, "QUIT" },
	{ SIGILL, "ILL" },
	{ SIGTRAP, "TRAP" },
	{ SIGABRT, "ABRT" },
	{ SIGIOT, "IOT" },
	{ SIGBUS, "BUS" },
	{ SIGFPE, "FPE" },
	{ SIGKILL, "KILL" },
	{ SIGUSR1, "USR1" },
	{ SIGSEGV, "SEGV" },
	{ SIGUSR2, "USR2" },
	{ SIGPIPE, "PIPE" },
	{ SIGALRM, "ALRM" },
	{ SIGTERM, "TERM" },
#ifdef SIGSTKFLT
	{ SIGSTKFLT, "STKFLT" },
#endif
	{ SIGSYS, "SYS" },
	{ SIGCHLD, "CHLD" },
	{ SIGCONT, "CONT" },
	{ SIGSTOP, "STOP" },
	{ SIGTSTP, "TSTP" },
	{ SIGTTIN, "TTIN" },
	{ SIGTTOU, "TTOU" },
	{ SIGURG, "URG" },
	{ SIGXCPU, "XCPU" },
	{ SIGXFSZ, "XFSZ" },
	{ SIGVTALRM, "VTALRM" },
	{ SIGPROF, "PROF" },
	{ SIGWINCH, "WINCH" },
	{ SIGIO, "IO" },
	{ SIGPOLL, "POLL" },
	{ SIGPWR, "PWR" }
};

void
do_kill(int argc, const char *const *argv)
{
	const char *	cp;
	int		sig = SIGTERM, pid;

	if ( *argv[1] == '-' )
	{
		unsigned	i = 0;
		sig = 0;
		cp = &argv[1][1];
		for ( ; i < sizeof(sig_desc) / sizeof(*sig_desc); ++i )
		{
			if ( !strcmp( cp, sig_desc[i].name ) )
			{
				sig = sig_desc[i].no;
				break;
			}
		}

		if ( !sig )
		{
			while (isDecimal(*cp))
				sig = sig * 10 + *cp++ - '0';

			if ( sig <= 0 || sig > 31 || *cp )
			{
				errmsg( "kill: unknown signal: %s", &argv[1][1] );
				return;
			}
		}

		--argc;
		++argv;
	}

	while (argc-- > 1)
	{
		cp = *(++argv);
		pid = 0;

		while (isDecimal(*cp))
			pid = pid * 10 + *cp++ - '0';

		if (*cp)
		{
			errmsg( "kill: non-numeric pid: %s", *argv );
			return;
		}

		if (kill(pid, sig) < 0)
			errmsg( "kill: %d: %s", pid, strerror(errno) );
	}
}


void
do_where(int argc, const char *const *argv)
{
	const char *	program;
	const char *	dirName;
	char *		path;
	char *		endPath;
	char *		fullPath;
	BOOL		found;

	found = FALSE;
	program = argv[1];

	if ( strchr(program, '/') )
	{
		errmsg( "where: program name cannot include a path" );
		return;
	}

	path = getenv("PATH");

	fullPath = getChunk(strlen(path) + strlen(program) + 2);
	path = chunkstrdup(path);

	if ( !path || !fullPath )
	{
		errmsg( "where: memory allocation failure: %s", strerror(errno) );
		return;
	}

	/*
	 * Check out each path to see if the program exists and is
	 * executable in that path.
	 */
	for (; path; path = endPath)
	{
		/*
		 * Find the end of the next path and NULL terminate
		 * it if necessary.
		 */
		endPath = strchr(path, ':');

		if (endPath)
			*endPath++ = '\0';

		/*
		 * Get the directory name, defaulting it to DOT if
		 * it is null.
		 */
		dirName = path;

		if (dirName == '\0')
			dirName = ".";

		/*
		 * Construct the full path of the program.
		 */
		strcpy(fullPath, dirName);
		strcat(fullPath, "/");
		strcat(fullPath, program);

		/*
		 * See if the program exists and is executable.
		 */
		if (access(fullPath, X_OK) < 0)
		{
			if (errno != ENOENT)
				printf("%s: %s\n", fullPath, strerror(errno));

			continue;
		}

		printf("%s\n", fullPath);
		found = TRUE;
	}

	if (!found)
		printf("Program \"%s\" not found in PATH\n", program);
}

void	do_losetup(int argc, const char *const *argv)
{
	int loopfd, targfd;
	struct loop_info loopInfo;

	if ( !strcmp( argv[1], "-d" ) )
	{
		loopfd = open( argv[2], O_RDONLY );
		if ( loopfd < 0 )
		{
			errmsg( "losetup: error opening %s: %s", argv[2], strerror(errno) );
			return;
		}

		if ( ioctl(loopfd, LOOP_CLR_FD, 0) < 0 )
		{
			errmsg( "losetup: error unassociating device: %s", strerror(errno) );
			close( loopfd );
			return;
		}

		close( loopfd );
		return;
	}

	loopfd = open( argv[1], O_RDONLY );
	if ( loopfd < 0 )
	{
		errmsg( "losetup: error opening %s: %s", argv[1], strerror(errno) );
		return;
	}

	targfd = open( argv[2], O_RDONLY );
	if ( targfd < 0 )
	{
		errmsg( "losetup: error opening %s: %s", argv[2], strerror(errno) );
		close( loopfd );
		return;
	}

	if ( ioctl(loopfd, LOOP_SET_FD, targfd) < 0 )
	{
		errmsg( "losetup: error setting up loopback device: %s", strerror(errno) );
		close( loopfd );
		close( targfd );
		return;
	}

	memset(&loopInfo, 0, sizeof(loopInfo));
	strcpy(loopInfo.lo_name, argv[2]);

	if ( ioctl(loopfd, LOOP_SET_STATUS, &loopInfo) < 0 )
		errmsg( "losetup: error setting up loopback device: %s", strerror(errno) );

	close( loopfd );
	close( targfd );

	return;
}

/* END CODE */
